import path from 'path';
import fs from 'fs';
import webpack, { DefinePlugin } from 'webpack';
import HtmlWebpackPlugin from 'html-webpack-plugin';
import MiniCssExtractPlugin from 'mini-css-extract-plugin';
import CopyWebpackPlugin from 'copy-webpack-plugin';
import ForkTsCheckerWebpackPlugin from 'fork-ts-checker-webpack-plugin';
import {RcbConfig} from './getConfig';
import utils from './utils';

function hasTSConfig() {
    return fs.existsSync(path.resolve(process.cwd(), 'tsconfig.json'));
}

function addPlugins(basePlugins: webpack.ResolvePlugin[], plugins?: webpack.ResolvePlugin[]) {
    if (plugins) {
        plugins.forEach(function (plugin: HtmlWebpackPlugin | DefinePlugin) {
            basePlugins.push(plugin);
        });
    }

    return basePlugins;
}

export default function getPlugins(rcb: RcbConfig) {
    let htmlWebpackPluginConfig;
    if (rcb.webpackPluginConfig) {
        htmlWebpackPluginConfig = rcb.webpackPluginConfig;
    }
    else {
        htmlWebpackPluginConfig = {
            template: path.resolve('./public/index.html')
        }
    }
    const basePlugins: webpack.ResolvePlugin[] = [
        new CopyWebpackPlugin({
            patterns: [{
                from: path.resolve('public'),
                globOptions: {ignore: ['**/index.html']}
            }]
        }),
        new HtmlWebpackPlugin(htmlWebpackPluginConfig),
        new webpack.DefinePlugin({
            'process.env': {
                NODE_ENV: JSON.stringify(utils.getEnv())
            }
        })
    ];

    if (hasTSConfig()) {
        basePlugins.push(new ForkTsCheckerWebpackPlugin({async: false}));
    }

    if (utils.isDev()) {
        addPlugins(basePlugins, [
            new webpack.HotModuleReplacementPlugin(),
            new MiniCssExtractPlugin({
                filename: 'css/[name].css'
            })
        ]);
    }

    if (utils.isProduction()) {
        addPlugins(basePlugins, [
            new MiniCssExtractPlugin({
                filename: 'css/[name].[hash:8].css'
            })
        ]);
    }

    return basePlugins;
};
