import fs from 'fs';
import webpack from 'webpack';
import webpackDevServer from 'webpack-dev-server';
import {Application} from 'express';
import {merge} from 'webpack-merge';
import {Options} from './start';
import PnpWebpackPlugin from 'pnp-webpack-plugin';

import getOutput from './getOutput';
import getLoaders from './getLoaders';
import getPlugins from './getPlugins';

import getOptimization from './getOptimization';
import mockMiddleware from './middleware/mock';
import loggerMiddleware from './middleware/logger';
import {RcbConfig} from './getConfig';
import utils from './utils';
import paths from './paths';

const entry = ['./src/index'];

export interface Configuration extends webpack.Configuration {
    devServer?: webpackDevServer.Configuration
}

function getExtraWebpackConfig(webpackConfig: Configuration) {
    let path: string = '';
    if (utils.isDev()) {
        path = paths.getResolvePath('webpack.config.dev.js');
    }

    if (utils.isProduction()) {
        path = paths.getResolvePath('webpack.config.prod.js');
    }

    if (!(path && fs.existsSync(path))){
        const webpackConfig = paths.getResolvePath('webpack.config.js');
        
        if (webpackConfig && fs.existsSync(webpackConfig)) {
            path = webpackConfig;
        }
    }

    if (!path) {
        return {};
    }

    try {
        const config = utils.requireModule(path);
        if (typeof config === 'function') {
            return config(webpackConfig);
        }
        return config;
    }
    catch (e) {
        return {};
    }
}

const HOST = '127.0.0.1';
const PORT = 3000;

export default (rcb: RcbConfig, options?: Options) => {
    const devServerConfig: webpackDevServer.Configuration = {
        host: rcb.host || HOST,
        port: rcb.port || PORT,
        proxy: rcb.proxy,
        quiet: true,
        overlay: true,
        disableHostCheck: true,
        historyApiFallback: {
            htmlAcceptHeaders: ['text/html', 'application/xhtml+xml', 'application/xml']
        },
        compress: true,
        clientLogLevel: 'warn',
        hot: true,
        inline: true,
        watchOptions: {
            ignored: /node_modules/
        },
        open: rcb.open === false ? rcb.open : true,
        setup(app: Application) {
            const mockConfig = rcb.mock;
            if (options?.mock && mockConfig) {
                // mock代理的url不会在node中log出来，只有proxy代理的url才会log方便查看是哪里出了问题
                app.use(mockConfig.api, mockMiddleware.middleware(rcb.mock));
            }
            app.use(loggerMiddleware());
        }
    };
    const defaultWebpackConfig: Configuration = {
        mode: utils.getEnv(),
        performance: {
            hints: false
        },
        devServer: utils.isDev() ? devServerConfig : undefined,
        devtool: utils.isDev() ? 'cheap-module-eval-source-map' : false,
        entry: entry,
        output: getOutput(),
        module: {
            rules: getLoaders(rcb)
        },
        optimization: getOptimization(),
        plugins: getPlugins(rcb),
        resolve: {
            alias: rcb.alias,
            plugins: [PnpWebpackPlugin],
            extensions: ['.js', '.jsx', '.es6', '.json', '.ts', '.tsx']
        },
        resolveLoader: {
            plugins: [
                PnpWebpackPlugin.moduleLoader(module)
            ]
        }
    };

    return merge(
        defaultWebpackConfig,
        rcb.webpack || {},
        getExtraWebpackConfig(defaultWebpackConfig)
    );
};
