import fs from 'fs';
import {Request, Response, NextFunction} from 'express';
import path from 'path';
import {match} from 'path-to-regexp';

import paths from '../paths';
import utils from '../utils';
import {MockConfig} from '../getConfig';

function readMockFile(req: Request, res: Response, next: NextFunction, mockConfig?: MockConfig,) {
    if (!mockConfig) {
        return new Error('没有找到mock配置');
    }
    try {
        const joinPath = paths.getJoinPath(mockConfig.rootDir, mockConfig.api, req.path);
        const reqUrl = path.join(mockConfig.api, req.url);
        if (
            mockConfig.whiteList
            && Array.isArray(mockConfig.whiteList)
            && mockConfig.whiteList.indexOf(reqUrl) > -1
        ) {
            const isWhite = mockConfig.whiteList.some(white => {
                const matchExp = match(white);
                return matchExp(reqUrl);
            });

            if (isWhite) {
                return next();
            }
        }

        const mock = getMockResponse(joinPath);
        if (utils.isFunction(mock)) {
            const ret = mock(req, res);
            // 如果返回的是object那么自动处理这个函数 如果函数内处理了就不要返回值
            if (utils.isObject(ret)) {
                res.send(ret);
            }
        }
        else if (mock !== undefined && mock !== false) {
            res.send(mock);
        }
        else {
            next();
        }
    }
    catch (e) {
        utils.error(e);
    }
}

function readJsFile(filePath: string) {
    try {
        const parentDir = path.resolve(filePath, '../');
        const files = fs.readdirSync(parentDir);

        return files
            .filter(filename => /\.js$/.test(filename))
            .map(filename => path.resolve(parentDir, filename));
    }
    catch (e) {
        return [];
    }
}

function getMockResponse(joinPath: string) {
    let mock;
    try {
        clearRequireCache(`${joinPath}.js`);
        mock = utils.requireModule(`${joinPath}.js`);
    }
    catch (e) {
        // 动态路由 例如 /test/1 => /test/:id
        const jsFiles = readJsFile(joinPath);
        let i = 0;
        let fileIndex = -1;
        const len = jsFiles.length;
        while (i < len) {
            const jsFile = jsFiles[i];
            const matchExp = match(jsFile.replace(/\.js$/, ''), {sensitive: true});
            const realFileInfo = matchExp(joinPath);

            if (realFileInfo) {
                fileIndex = i;
                break;
            }

            i++;
        }

        if (fileIndex > -1) {
            clearRequireCache(jsFiles[fileIndex]);
            mock = require(jsFiles[fileIndex]); // 异步过程
        }
    }

    return mock;
}

function clearRequireCache(filePath: string) {
    const cacheMod = require.cache[filePath];

    // 防止内存泄漏
    if (cacheMod && cacheMod.parent) {
        const index = cacheMod.parent.children.indexOf(cacheMod);

        if (index > -1) {
            cacheMod.parent.children.splice(index, 1);
        }
    }

    delete require.cache[filePath];
}

export default {
    middleware: function (options?: MockConfig) {
        return function (req: Request, res: Response, next: NextFunction) {
            readMockFile(req, res, next, options);
        };
    }
};
